/*  Intelligent FRAC.		(C) 2000  Michael Glickman      */
/*  --------------------------------------------------- */
/*  See LICENSE regarding distribution policy and       */
/*  conditions of use.                                  */ 
/*  =================================================== */

#include <X11/Xlib.h>
#include <X11/extensions/xf86dga.h>
#include <X11/Xatom.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <pwd.h>

#include "ifr.h"
#ifdef USE_VM_EXTENSION
#include <X11/extensions/xf86vmode.h>
#endif

#include "ifr_pixmap.h"

#include "ifr_graph_common.h"

extern  PLAYER *players[];
extern  GAME	game;
PLATFORM pltf = PLTF_DGA;
CMAP_TYPE cm_type;

typedef unsigned char *BITMAP;
typedef unsigned long  COLOUR;

extern int	PrivateColours;
extern  int LastSaveLayer;

static	int 	alloc_colour_count = 0;
static	int 	xf86release = 0;

int BytesPerPixel;


Display *disp;
Window root_wnd, wnd;
GC	gc;
static	Visual  *visual;
static  Colormap cm, cm_original;
static char *FBaddr;
static int  FBlwid, FBbanksize, FBmemsize, FBCurBank;
static int  screen_no;
static int  WndLeft, WndTop;

static int dga_flags = 0;
#define DGA_DIRECT_VIDEO_ON  1
#define DGA_VM_CHANGED	     2
#define DGA_CM_CHANGED	     4	

static int alloc_colour_space(void);
static int alloc_colour_cells(void);

static const char *WindowIdAtomName = "DIFRAC_WINDOW_ID";
static Atom WindowIdAtom = None;

typedef struct 
{	int width;
	int height;
	char *pixels;
	int  start_colour;
	char end_colour;
} PIXMAP_INFO;

#define StippleCount 3
#define PaneCount 3
BITMAP all_stipples[StippleCount];
PIXMAP_INFO  *all_panes[PaneCount];


unsigned char *cm_pixels = NULL;

#ifdef USE_VM_EXTENSION
static XF86VidModeModeInfo modeline_original;
static XF86VidModeModeInfo modeline_current;
extern int PreserveVM;
#endif

static pid_t find_bossed_process(void);


PIXMAP_INFO *create_pixmap(int width, int height)
{	PIXMAP_INFO *pi;

	pi = (PIXMAP_INFO *)malloc(sizeof(PIXMAP_INFO));
	if (!pi) return NULL;

	pi->pixels = malloc(width * height * BytesPerPixel);
	if (pi->pixels == NULL)
	{ free(pi); return NULL; }

	pi->width = width;
	pi->height = height;
	return pi;
}

void free_pixmap(PIXMAP_INFO *pi)
{
	free(pi->pixels);
	free(pi);
}

/* ================================================================= */
static unsigned short extend_colour_component(unsigned int c)
{
    unsigned short result;
    
    result = c << 8;
    if ((c & 0xf) == 0xf) result |= 0xff;
    
    return result; 
}  

static int alloc_colour_space(void)
{
    size_t space_used;
	if (ScrDepth <= 4) space_used = 17;
			else    space_used = alloc_colour_count;

    space_used *= BytesPerPixel;
	cm_pixels = (unsigned char *) malloc(space_used);
    return (cm_pixels != NULL);
	
}

static int alloc_colour_cells(void)
{
	int rc, i;
	COLOUR *xcolours;
	
	xcolours = (COLOUR *) malloc(alloc_colour_count * sizeof(COLOUR));  
	if (xcolours == NULL) return 0;
	rc = XAllocColorCells(disp, cm, False,
	       NULL, 0, xcolours, alloc_colour_count);
	if (rc)
	{  for (i=0; i<alloc_colour_count; i++)
	     cm_pixels[i] = (unsigned char) xcolours[i];
	}
	free (xcolours);			   
	return rc;	  		   
}    

static int set_static_stored_colours(int start_colour, int colour_count,
        const unsigned char rgb[])
{	int i;
    int *pal_buffer, *pal_buf_ptr;
	const unsigned char *rgbptr;
	XColor  xc;

	pal_buf_ptr = pal_buffer = (int *) malloc(StaticColours * 3 * sizeof(int));
	if (pal_buffer == NULL) return 0;
	rgbptr = rgb;
	    
    for (i=0; i<colour_count; i++)
	{
	   memset(&xc, '\0', sizeof(xc));
	   xc.red =   extend_colour_component(*rgbptr++); 
	   xc.green = extend_colour_component(*rgbptr++);  
	   xc.blue =  extend_colour_component(*rgbptr++); 
	   xc.pixel = (COLOUR) cm_pixels[start_colour+i];
	   xc.flags = DoRed | DoGreen | DoBlue;	   
	   if (!XStoreColor(disp, cm, &xc)) break;
	}	   

	i = (i >= colour_count);
	if(i) dga_flags |= DGA_CM_CHANGED;
	return  i;

}

static int set_static_high_colours(int start_colour, int colour_count,
        const unsigned char rgb[])
{	int i, s;    
    unsigned char *pal_buf_ptr;
	const unsigned char *rgbptr;
	union { unsigned int s; unsigned char c[2]; } pix_union;
	unsigned short r, g, b;

	pal_buf_ptr = cm_pixels + start_colour*2;
	s = ScrDepth - 15; rgbptr = rgb;
    
  for (i=0; i<colour_count; i++)
	{	 r = *rgbptr++ >> 3;
		 g = *rgbptr++ >> (3-s);
		 b = *rgbptr++ >> 3;
		 pix_union.s = (r << (10+s)) | (g << 5) | b;
		 *pal_buf_ptr++ = pix_union.c[0];
		 *pal_buf_ptr++ = pix_union.c[1];
	}
  return  1;

}

static int set_static_true_colours(int start_colour, int colour_count,
        const unsigned char rgb[])
{	int i, s;    
	unsigned char *pal_buf_ptr;
	const unsigned char *rgbptr;

	pal_buf_ptr = cm_pixels + start_colour*BytesPerPixel;
	rgbptr = rgb; s = BytesPerPixel-3;

	for (i=0; i<colour_count; i++)
	{		*pal_buf_ptr++ = rgbptr[2];
			*pal_buf_ptr++ = rgbptr[1];
			*pal_buf_ptr++ = rgbptr[0];
			rgbptr += 3;
			if (s>0)
			{	memset(pal_buf_ptr, '\0', s); pal_buf_ptr += s; }
	}

  return  1;
}

int set_static_colours(int start_colour, int colour_count,
        const unsigned char rgb[])
{	switch (BytesPerPixel)
	{  case 1:
			return set_static_stored_colours(start_colour, colour_count, rgb);
 
	   case 2:	 	
			return set_static_high_colours(start_colour, colour_count, rgb);

	   default:	 	
			return set_static_true_colours(start_colour, colour_count, rgb);
	}


}

/* ---------------------------------------------------------------- */
int set_bkgr_stored_colours(const unsigned short *base_colour,
                            const unsigned char *grey_levels, 
                          	int colour_start, int colours_used)
{ 
    unsigned short r, g, b, grey;
	XColor xc;
    int x;


	r = base_colour[0];
	g = base_colour[1];
	b = base_colour[2];
	
    for (x = 0; x<colours_used; x++)
    {   grey = grey_levels[x];
    	xc.red =   grey * r;
		xc.green = grey * g;
		xc.blue =  grey * b;
		if (colour_start >= alloc_colour_count) break;
		xc.pixel = (COLOUR) cm_pixels[colour_start+x];
		xc.flags = DoRed | DoGreen | DoBlue;	   
	if (!XStoreColor(disp, cm, &xc)) break;
    }

    x = (x >= colours_used);
    if (x) dga_flags |= DGA_CM_CHANGED;

    return  x;
}

int set_bkgr_high_colours(const unsigned short *base_colour,
                            const unsigned char *grey_levels, 
                          	int colour_start, int colours_used)
{ 
    unsigned short r, g, b, r1, g1, b1, grey;
    unsigned char *pal_buf_ptr;
		union { unsigned int s; unsigned char c[2]; } pix_union;
    int x, s;

    pal_buf_ptr = cm_pixels + colour_start*2;

	r = base_colour[0];
	g = base_colour[1];
	b = base_colour[2];
	s = ScrDepth-15;

    for (x = 0; x<colours_used; x++)
    {   grey = grey_levels[x];
		r1 = (grey * r) >> 11;	
		g1 = (grey * g) >> (11-s);	
		b1 = (grey * b) >> 11;	
		pix_union.s = (r1 << (10+s)) | (g1 << 5) | b1;
		*pal_buf_ptr++ = pix_union.c[0];				
		*pal_buf_ptr++ = pix_union.c[1];				
    }

  	return 1;
}

int set_bkgr_true_colours(const unsigned short *base_colour,
                            const unsigned char *grey_levels, 
                          	int colour_start, int colours_used)
{ 
    unsigned short r, g, b, grey;
    unsigned char *pal_buf_ptr;
    int x, s;

    pal_buf_ptr = cm_pixels + colour_start*BytesPerPixel;

	r = base_colour[0];
	g = base_colour[1];
	b = base_colour[2];
	s = BytesPerPixel-3;

    for (x = 0; x<colours_used; x++)
    {   grey = grey_levels[x];
		*pal_buf_ptr++ = (unsigned char)((grey * b) >> 8);	
		*pal_buf_ptr++ = (unsigned char)((grey * g) >> 8);	
		*pal_buf_ptr++ = (unsigned char)((grey * r) >> 8);	
		if (s>0) { memset(pal_buf_ptr, '\0', s); pal_buf_ptr += s; }
	}
  	return 1;
}

int set_bkgr_colours(const MPMP_INFO *mpmpi, 
					 const unsigned short *base_colour,	int colour_start)
{
  int colours_used;
  unsigned char *grey_levels;    

  colours_used = mpmpi->colours_used;
  grey_levels = mpmpi->grey_levels;    

  switch (BytesPerPixel)
  {  case 1:
		return set_bkgr_stored_colours(base_colour,grey_levels, 
                          	       colour_start, colours_used);
     case 2:	 	
		return set_bkgr_high_colours(base_colour,grey_levels, 
                          	       colour_start, colours_used);
     default:	 	
		return set_bkgr_true_colours(base_colour,grey_levels, 
                          	       colour_start, colours_used);
  }
}
/* ================================================================= */

void graph_end(void)
{
    int i;
    PIXMAP_INFO *xpm;

    if(dga_flags & DGA_DIRECT_VIDEO_ON) 
	  XF86DGADirectVideo(disp, screen_no, 0);

#ifdef USE_VM_EXTENSION
    if(dga_flags & DGA_VM_CHANGED)
	{  XF86VidModeLockModeSwitch(disp, screen_no, 0);
	   XF86VidModeSwitchToMode(disp, screen_no, &modeline_original);
	}	   
#endif

	if (disp != NULL)
	{
		if (WindowIdAtom != None)  	
		{ pid_t boss_pid = find_bossed_process();
		  if (boss_pid == getpid()) 
			XDeleteProperty(disp, root_wnd, WindowIdAtom);
		}			
		
	
		if (PrivateColours)
		{ XSetWindowColormap(disp, wnd, cm_original);
		  XFreeColormap(disp, cm);
		}	  
		XCloseDisplay(disp);
	}		
	


	if (cm_pixels) free(cm_pixels);

    for (i=0; i<PaneCount; i++)
    { xpm = all_panes[i];
      if (xpm != NULL)  free_pixmap(xpm);    
    }



}


#ifdef USE_VM_EXTENSION
void copy_video_info(XF86VidModeModeInfo *to,
                     const XF86VidModeModeInfo *from)
{ memcpy(to, from, sizeof(XF86VidModeModeInfo));
  to->privsize = 0;
  to->private = NULL;
}
  
int switch_video_mode(int NewWidth, int NewHeight)
{
  int event_base,error_base;
  int modelines_count, i;
  XF86VidModeModeInfo **modelines;
  
	  
  if (XF86VidModeQueryExtension(disp, &event_base, &error_base) == False)
  {  fprintf (stderr, "Sorry, this X configuration does appear to support XF86VidMode: running in window\n");
     fprintf (stderr, "Possible solution: comment out line with \"DisableVidModeExtension\" in XF86Config\n");
     return 0;   
  }
  	  
  if (XF86VidModeGetAllModeLines(disp, screen_no, &modelines_count, &modelines) == False ||
                 modelines_count <= 0)
  {  fprintf (stderr, "Could not get list of video modes: running in window.\n");
     return 0;   
  }

  copy_video_info(&modeline_original, modelines[0]);
  ScrWidth = modelines[0]->hdisplay;
  ScrHeight = modelines[0]->vdisplay;

  if (PreserveVM || (ScrWidth == NewWidth && ScrHeight == NewHeight))
  {  i = 1;  goto Leave; }
  
  XF86VidModeLockModeSwitch(disp, screen_no, 0);
  for (i=1; i<modelines_count; i++)
  {	if(modelines[i]->hdisplay == NewWidth && modelines[i]->vdisplay == NewHeight &&
      XF86VidModeSwitchToMode(disp, screen_no, modelines[i]) == True)
		{  ScrWidth = NewWidth; ScrHeight = NewHeight;
		   XF86VidModeLockModeSwitch(disp, screen_no, 1);
		   copy_video_info(&modeline_current, modelines[i]);
	  	   dga_flags |= DGA_VM_CHANGED;
		   break;
		}	   
  }


  if (i>= modelines_count)
  {  fprintf (stderr, "Could not find appropriate video mode: running in window.\n");
     fprintf (stderr, "Hint: video mode %dx%d should be listed in \"Modes\" line\n", NewWidth, NewHeight);
	 fprintf (stderr, "for the current depth: use xf86config, or add mode to\n");
	 fprintf (stderr, "XF86Config: \"Screen\"/\"Display\"/\"Modes\"\n");
     fprintf (stderr, "See 'Using XFree86 extensions' in ifrac INSTALL for more hints.\n");
     i = 0;
  }
  else
	 i = 1;

Leave:	 
  { int x, y;
	 XF86VidModeGetViewPort(disp, screen_no, &x, &y);
	 if (x!=0 || y!=0)	
	   XF86VidModeSetViewPort(disp, screen_no, 0, 0);
  }		   
  XFree((void *) modelines);	 	 
  return i;

}


#endif



int set_video_mode(void)
{
    int event_base,error_base;
    int flags;
  
	if (geteuid() != 0)
	{  fprintf (stderr, "Warning: you need 'suid root' to run difrac:\n"
						"Log in as root and run 'make install'!\n");	
	}


    if (XF86DGAQueryExtension(disp, &event_base, &error_base) == False)
    {  fprintf (stderr, "Sorry, this X configuration does appear to support XF86DGA ! \n");
	   if (xf86release < 3300)
  	     fprintf (stderr, "You might need to upgrade XFree86 server\n");
	   else
	   if (xf86release >= 3400)
	      fprintf (stderr, "Possible solution: disable \"omit xfree86-dga\" in XF86Config.\n");

       fprintf (stderr, "See 'Using XFree86 extensions' in ifrac INSTALL for more hints.\n");
	   return 0;   
    }

    XF86DGAQueryDirectVideo(disp, screen_no, &flags);
    if ((flags & XF86DGADirectPresent) == 0)
    {  fprintf (stderr, "Sorry, this XF86DGA version does appear to support direct graphics! \n");
       return 0;   
    }

    if (XF86DGAGetVideo(disp, screen_no, &FBaddr, &FBlwid, &FBbanksize, &FBmemsize) == False)
    {  fprintf (stderr, "Failed to get video memory attributes\n");
       return 0;   
    }

    if (XF86DGAGetVideo(disp, screen_no, &FBaddr, &FBlwid, &FBbanksize, &FBmemsize) == False)
    {  fprintf (stderr, "Failed to get video memory attributes\n");
       return 0;   
    }

	switch(VideoRes)
	{	
		case VRES_NORMAL:
	    WndWidth = 640;		    
	    WndHeight = 480;		    
	    break;

		case VRES_LOW:
	    WndWidth = 600;		    
	    WndHeight = 400;		    
	    break;

		case VRES_DEFAULT:
	    WndWidth = (ScrWidth <= 640) ? 600 : 640;
	    WndHeight = (ScrHeight <= 480) ? 400 : 480;
		break;
	}						

#ifdef USE_VM_EXTENSION
   switch_video_mode(WndWidth, WndHeight);
#endif

	if (WndWidth > ScrWidth || WndHeight > ScrHeight)
	{ fprintf (stderr, "Insufficient screen resolution: "
                     "%dx%d is required\n", WndWidth, WndHeight);
	  return 0;
	}
	
	WndLeft = (ScrWidth - WndWidth) >> 1;
	WndTop = (ScrHeight - WndHeight) >> 1;

	return 1;

}

int graph_init_video(void)
{
    Screen	*scr;
	int	vis_class;
  
	dga_flags = 0;  
    disp = XOpenDisplay(NULL);
    pltf = PLTF_DGA;
    if (disp == NULL)
    {  fprintf (stderr, "Can't open display: this is an X-Window application! \n");
       return 0;   
    }

    WindowIdAtom = XInternAtom(disp, WindowIdAtomName,  False); 
	xf86release = VendorRelease(disp);
	
    scr = DefaultScreenOfDisplay(disp);
    root_wnd = scr->root;
    ScrWidth = scr->width;
    ScrHeight = scr->height;
    cm = cm_original = scr->cmap;
    visual = scr->root_visual;
	gc = scr->default_gc;
	vis_class = visual->class;
  
	screen_no = DefaultScreen(disp);

    ScrDepth = scr->root_depth; /* scr->root_visual->bits_per_rgb; */
    if (ScrDepth < 4)
    {
		fprintf(stderr, "Sorry, at least 16 colours are needed to run the application");
		return 0;	    
    }

    BytesPerPixel = (ScrDepth+7)/8;

	switch(vis_class)
    {
		case TrueColor:
	    	cm_type = CM_TRUE;
		   break;
	    
		case PseudoColor:
		    cm_type = CM_PALETTE;
		    break;

		default:
		    cm_type = CM_STATIC;
    }	    

	if (BytesPerPixel > 1  && cm_type != CM_TRUE)
    {  fprintf (stderr, "Sorry, only true colour mode is supported with screen depth > 8.\n");
       return 0;   
    }

	stored_colours = (cm_type == CM_PALETTE);

	if (!stored_colours) PrivateColours = 0;

    return set_video_mode();

}

int graph_start(void)
{
	int i;
    PIXMAP_INFO  *xpm;    
	unsigned char   bw_colours[6];

	
    if (BytesPerPixel > 1) i = 256; else i = 1 << ScrDepth;
    cm_pixels = (unsigned char *) malloc(i * BytesPerPixel);
    if (cm_pixels == NULL) return 0;

    if (XF86DGADirectVideo(disp, screen_no, XF86DGADirectGraphics) == False)
    {  fprintf (stderr, "Failed to set direct video mode\n");
       return 0;   
    }

    FBCurBank = 0;
    XF86DGASetVidPage(disp, screen_no, FBCurBank);
    
    dga_flags |= DGA_DIRECT_VIDEO_ON;
    wnd = root_wnd;

    /* ------------- Colour allocation ------------ */
	if (ScrDepth <= 4)
	    alloc_colour_count = 16;
	else
	{ 	alloc_colour_count = StaticColours + mandel_mpmp_info.colours_used;
		i = StaticColours + hypoth_mpmp_info.colours_used;

		if (i > alloc_colour_count) alloc_colour_count = i;
		i = 1 << ScrDepth;
		if (i < alloc_colour_count) alloc_colour_count = i;
	}		

    if (!alloc_colour_space()) return 0;

    if (cm_type == CM_PALETTE && PrivateColours == 0)
    {
       if (!alloc_colour_cells())
       { fprintf(stderr, "difrac: couldn't allocate all colours; private colour map will be used\n");
          PrivateColours = 1;
       }	
    }
    	
    /* Creating a colour map */
    if (PrivateColours) 
    {
	    cm = XCreateColormap(disp, root_wnd, visual, AllocNone);
        if (cm == None) return 0;

        if (cm_type == CM_PALETTE && 
	  	    !alloc_colour_cells()) return 0;
		 XSetWindowColormap(disp, wnd, cm);

    }

    for (i=0; i<PaneCount; i++)
		all_panes[i] = NULL;

    for(i=0; i<3; i++)
    {  bw_colours[i] = 0;
       bw_colours[i+3] = 255;
    }

    if (!set_static_colours(0, 2, bw_colours))
				return 0; 

    all_stipples[ST_NO_STIPPLE] = NULL;
    all_stipples[ST_DARK_STIPPLE] = quarter_bmp_bits;
    all_stipples[ST_LIGHT_STIPPLE] = half_bmp_bits;

    for (i=0; i<PaneCount; i++)
    {	xpm = create_pixmap(WndWidth, WndHeight);   
		if(xpm == NULL) break;
		all_panes[i] = xpm;
    }	  

    return (i>=PaneCount);
          	
}


void  dga_drawscansegment(char *image, int x, int y, int wid)
{
    int offset, bank;
    int	w1, w2;
    
    offset = ((y+WndTop) * FBlwid + x + WndLeft) * BytesPerPixel;
    bank = offset / FBbanksize;
    offset %= FBbanksize;
    
    if (bank != FBCurBank)
    { if( XF86DGASetVidPage(disp, screen_no, bank) == False) return;
      FBCurBank = bank;     
    }

    w1 = FBbanksize - offset;
    if (w1 >= wid) { w1 = wid; w2 = 0; }
	     else  w2 = wid - w1;   
	     
    memcpy(FBaddr + offset, image, w1);
    
/*    if (w2 > 0 && XF86DGASetVidPage(disp, screen_no, ++bank ) == True) */

    if (w2 > 0)
    {  bank++;
       if(XF86DGASetVidPage(disp, screen_no, bank ) == True) 
		{  FBCurBank = bank;
           memcpy(FBaddr, image+w1, w2);      		            
        }
    }	
          
}

void get_graphics_line(char *value)
{
#ifdef USE_VM_EXTENSION
    if(dga_flags & DGA_VM_CHANGED)
	   snprintf (value, 81, "Screen resolution: %d x %d, clock %.2f, depth: %d",
											  ScrWidth, ScrHeight,
												modeline_current.dotclock/1000.,  ScrDepth);
	 
	 else
#endif
    snprintf (value, 81, "Screen resolution: %d x %d, depth: %d, window size: %d x %d",
											  ScrWidth, ScrHeight, ScrDepth,
												WndWidth, WndHeight);
}

void flush_pane(PANE_TYPE which_pane, int x, int y, int wid, int hei)
{
	char *image;
	int  width;
	PIXMAP_INFO *pixmap;

    if (dga_flags & DGA_CM_CHANGED)
    {
//	  XF86DGAInstallColormap(disp, screen_no, cm);
	  XInstallColormap(disp,  cm);
	  XSync(disp, False);
  	  dga_flags &= ~DGA_CM_CHANGED;
    }		    	

    pixmap = all_panes[which_pane];
    width = pixmap->width * BytesPerPixel;
    image = pixmap->pixels+ y*width+ x*BytesPerPixel;
    wid   *= BytesPerPixel;

    while (--hei>=0)
    {
	dga_drawscansegment(image, x, y, wid);
	image += width; y++; 				
    }
	

}

void flush_all_pane(PANE_TYPE which_pane)
{

    flush_pane(which_pane, 0, 0, WndWidth, WndHeight);

}

static void copy_pmp(PIXMAP_INFO *source, PIXMAP_INFO *destination, 	
          int srcx, int srcy, int wid, int hei, int destx, int desty)
{
	char *src, *dest;
	int srcwid, destwid;

	srcwid = source->width * BytesPerPixel;
	destwid = destination->width * BytesPerPixel;
	src  = source->pixels + srcy*srcwid + srcx*BytesPerPixel;
	dest = destination->pixels + desty*destwid + destx*BytesPerPixel;
	wid *= BytesPerPixel;
	
	while (--hei >= 0)
	{ memcpy(dest, src, wid);
	  src += srcwid;
	  dest += destwid;
	}
}

void copy_pane(PANE_TYPE pt_source, PANE_TYPE pt_destination, 	
          int srcx, int srcy, int wid, int hei, int destx, int desty)
{
	copy_pmp(all_panes[pt_source], all_panes[pt_destination],
				srcx, srcy, wid, hei, destx, desty);
}


void  draw_line(PANE_TYPE which_pane, int x, int y,
			     int wx, int wy, COLENTRY fillcol,
                 COLENTRY bgcol, STIPPLE_TYPE which_stipple)
{
	PIXMAP_INFO *pane;
	BITMAP  stipple;

	int swx, swy, count;
	int dxnum, dxden, dynum, dyden;
	int  width, x1, y1;
	unsigned char *image, *colptr, *imptr;
	COLENTRY  col;


	/* Get sign and abs values */
	swx = (wx >= 0) ? 1 : -1;	
	swy = (wy >= 0) ? 1 : -1;	
	wx *= swx;  wy *= swy;

	if (wx >= wy)
	{ /* Going along x */
		if (wx == 0) return;
		dxnum=swx; dxden = 1;
		dynum=wy*swy; count = dyden=wx;
	}
	else
	{ /* Going along y */
		dynum=swy; dyden = 1;
		dxnum=wx*swx; count = dxden=wy;
	}


	pane = all_panes[which_pane];
	stipple = all_stipples[which_stipple];

	width = pane->width;
	image = pane->pixels;
	count++;
	
	col = fillcol;
 
	x1 = x * dxden;
	y1 = y * dyden;
	while(--count>=0)
	{  x = x1/dxden; y = y1/dyden;		
	  if (stipple) 
	  	     col = stipple[y&1] & (1<<(x&1)) ? fillcol : bgcol;
  	  if(col >= 0 && x < width)
	  { imptr = image + (width*y + x) * BytesPerPixel;
		colptr = cm_pixels + col * BytesPerPixel;
		memcpy(imptr, colptr, BytesPerPixel);
		imptr += BytesPerPixel;
	  }
	  x1 += dxnum;  y1 += dynum;
		 	
	}

}

/* which_pane:  0 - draw in background, 1 - draw in foreground */
void  fill_hor_trapezium(PANE_TYPE which_pane, int x, int y,
	          int w, int h, int dw1, int dw2,
			  COLENTRY fillcol, COLENTRY bgcol, STIPPLE_TYPE which_stipple)
{

	int i;

	if (h==0) return;
	
	dw2 -= dw1;
	if (h<0) 
	{  x += dw1; y += h; w += dw2;
	   h = -h;  dw1 = -dw1; dw2 = -dw2;
	}	   
	   
	w *= h; x *= h;
		
	for (i=0; i<h; i++)
	{  draw_line(which_pane, x/h, y, w/h, 0, 
	       fillcol, bgcol, which_stipple);
	   x += dw1;  w += dw2;  y++;				  
	}
	

}			  

/* which_pane:  0 - draw in background, 1 - draw in foreground */

void  fill_parallelogramm(PANE_TYPE which_pane, int x, int y,
			        int wx, int wy, int hx, int hy,
				    COLENTRY fillcol, COLENTRY bgcol, STIPPLE_TYPE which_stipple)
{
	int swx, swy, count;
	int dxnum, dxden, dynum, dyden;

	/* Get sign and abs values */
	swx = (wx >= 0) ? 1 : -1;	
	swy = (wy >= 0) ? 1 : -1;	
	wx *= swx;  wy *= swy;

	if (wx >= wy)
	{ /* Going along x */
		dxnum=swx; dxden = 1;
		dynum=wy*swy; count = dyden=wx;
	}
	else
	{ /* Going along y */
		dynum=swy; dyden = 1;
		dxnum=wx*swx; count = dxden=wy;
	}

	x *= dxden;
	y *= dyden;
	while(--count>=0)
	{  draw_line(which_pane, x/dxden, y/dyden, hx, hy, fillcol, bgcol, which_stipple);
	   x += dxnum; y += dynum;
	}  	

}

int set_line_colour( unsigned short base_colour[3])
{
   unsigned char rgb[3];
	 int i;

	 for (i=0; i<3; i++)
			rgb[i] = (unsigned char) base_colour[i]; 
	 
	 return set_static_colours(line_colour, 1, rgb);
}


int set_bkgr_line_colours(const MPMP_INFO *mpmpi, unsigned short base_colour[3])
{
	return 
		set_bkgr_colours(mpmpi, base_colour, StaticColours) &&
		set_line_colour(base_colour);
}

/*
int insert_to_pixmap(int which_pane, const MPMP_INFO *mpmpi,
	      unsigned short  base_colour[3],
			int sx, int sy,	 int colour_start)
{
    PIXMAP_INFO  *xpm;    
	int  width, height;
    unsigned char *pixels,  *bkgr_xpm_ptr, p;    

	int  x, y;

    if (!set_bkgr_colours(mpmpi, base_colour, colour_start)) return 0;
    width = mpmpi->width;
    height = mpmpi->height;

	xpm = all_panes[which_pane];

    bkgr_xpm_ptr = xpm->pixels + sy*WndWidth + sx;
    pixels = mpmpi->pixels;
   
    for (y=0; y<height; y++)
    {
	   for (x=0; x<width; x++)	
	   {   p = *pixels++;
		   if (p!=0)  *bkgr_xpm_ptr = colour_start + p;  
		   bkgr_xpm_ptr++;
		}
	
		bkgr_xpm_ptr += (WndWidth-width);
    }
	return 1;
}
*/

int fill_background_pixmap(int which_pane, const MPMP_INFO *mpmpi,
      const unsigned short  *base_colour, int first_colour)
{
    PIXMAP_INFO  *xpm;    
    unsigned char *pixels,  *bkgr_xpm_ptr;    
		int	x, y, width, height;

	  if (base_colour &&
		 		!set_bkgr_colours(mpmpi, base_colour, first_colour)) return 0;

    width = mpmpi->width;
    height = mpmpi->height;

		xpm = all_panes[which_pane];

    bkgr_xpm_ptr = xpm->pixels;
    pixels = mpmpi->pixels + OffsetY*width + OffsetX;
   
    for (y=0; y<WndHeight; y++)
    {
		  for (x=0; x<WndWidth; x++)	
			{  memcpy(bkgr_xpm_ptr,
               cm_pixels+(first_colour+*pixels++)*BytesPerPixel,
               BytesPerPixel);  
//				*bkgr_xpm_ptr = cm_pixels[first_colour+*pixels++];
//				*bkgr_xpm_ptr = first_colour+*pixels++;
			  bkgr_xpm_ptr += BytesPerPixel;
			}

			pixels += (width - WndWidth);
    }

	return 1;
}
	      

ACTION morph_to_colour(unsigned short new_base_colour[3])
{
	int i, k;
  unsigned short mid_colour[3];
  unsigned long next_clock;
  ACTION new_action = ACT_NONE;

  for (k=1; k<=16; k++)
  {
    next_clock = start_clock(DRAW_LEVEL_DELAY2);
    for (i=0; i<3; i++)
	  mid_colour[i] = (current_base_colour[i] * (16-k) +
					  new_base_colour[i] * k) >> 4;
  	
		if (!set_bkgr_line_colours(&mandel_mpmp_info, mid_colour))
				  return new_action;
    new_action = wait_clock(next_clock, NULL);
		if (new_action != ACT_NONE) break;
  }

  set_bkgr_line_colours(&mandel_mpmp_info, new_base_colour);
  return new_action;  
}

ACTION morph_to_image(PANE_TYPE source_pane)
{
  int  x, y, i;
  unsigned long next_clock;
  ACTION new_action = ACT_NONE;
  PIXMAP_INFO *from, *to;
  unsigned char *ptrf, *ptrt;
  unsigned char *ptrf1, *ptrt1;
    
  from  = all_panes[source_pane];
  to  = all_panes[PT_FOREGROUND];
  i = MORPH_STEP;

  while (--i >= 0)
  {
    next_clock = start_clock(DRAW_LEVEL_DELAY1);

	for (y=0; y<WndHeight; y++)
	{	
	  ptrf = from->pixels + y*WndWidth*BytesPerPixel;
	  ptrt = to->pixels + y*WndWidth*BytesPerPixel;

	  for (x= (y*3+i) % MORPH_STEP; x<WndWidth; x += MORPH_STEP)
	  {	 ptrf1 = ptrf + x*BytesPerPixel;
		 ptrt1 = ptrt + x*BytesPerPixel;
		 memcpy(ptrt1, ptrf1, BytesPerPixel);
	  }
	}

    flush_pane(PT_FOREGROUND, 0, 0, WndWidth, WndHeight);
		new_action = wait_clock(next_clock, NULL);
    if (new_action != ACT_NONE) break;
  }			

  if (i>=0)
  {	copy_pane(source_pane, PT_FOREGROUND, 0, 0, WndWidth, WndHeight, 0, 0);
    flush_pane(PT_FOREGROUND, 0, 0, WndWidth, WndHeight);
  }	 

#ifdef USE_VM_EXTENSION
	 XF86VidModeGetViewPort(disp, screen_no, &x, &y);
#endif
  return new_action;
}

ACTION fade_image(COLENTRY colentry)
{
  int  x, y, i;
  unsigned long next_clock;
  ACTION new_action = ACT_NONE;
  PIXMAP_INFO  *to;
  unsigned char *ptrt, *ptrt1, *colptr;

	colptr = cm_pixels + colentry * BytesPerPixel;    
  to  = all_panes[PT_FOREGROUND];
  i = MORPH_STEP;

  while (--i >= 0)
  {
    next_clock = start_clock(DRAW_LEVEL_DELAY1);

		for (y=0; y<WndHeight; y++)
		{ 	ptrt = to->pixels + y*WndWidth*BytesPerPixel;

		  	for (x= (y*3+i) % MORPH_STEP; x<WndWidth; x += MORPH_STEP)
			  {	 ptrt1 = ptrt + x*BytesPerPixel;
					 memcpy(ptrt1, colptr, BytesPerPixel);
			  }
		}

    flush_pane(PT_FOREGROUND, 0, 0, WndWidth, WndHeight);
		new_action = wait_clock(next_clock, NULL);
    if (new_action != ACT_NONE) break;
  }			

#ifdef USE_VM_EXTENSION
	 XF86VidModeGetViewPort(disp, screen_no, &x, &y);
#endif
  return new_action;
}


/* ========================================================= */
static pid_t find_bossed_process(void)
{ 
    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *data;
	Window  pid_return;

	if (WindowIdAtom == None) return -1;
   
    XGetWindowProperty(disp, root_wnd, WindowIdAtom, 0,
			     1, True, XA_INTEGER,
			     &actual_type,  &actual_format,
			     &nitems, &bytes_after, &data); 
    
    if (actual_type == None || nitems == 0) return -1;
    
    pid_return = *(Window *)data;	    
    XFree(data);

	if (kill(pid_return, 0) < 0 &&
               errno != EPERM) return -1;	
	
    return pid_return;
}

/* Mode = 0 - kill, 1 - resume */
void release_bossed_process (pid_t pid, int mode)
{
    XEvent xevnt;
   

    xevnt.xclient.type = ClientMessage;
    xevnt.xclient.serial = 0;
    xevnt.xclient.send_event = True;
    xevnt.xclient.message_type = WindowIdAtom; 
    xevnt.xclient.format = 32;
    xevnt.xclient.data.l[0] = (long) pid;
    xevnt.xclient.data.l[1] = (long) mode;
    xevnt.xclient.display = disp;

    XSendEvent(disp, root_wnd, False, KeyPressMask, &xevnt);   
}

void boss_routine(void)
{   XEvent xevnt;
	pid_t my_pid;
#ifdef USE_BACKGROUND_MUSIC
	extern MUSIC_STATUS mstatus;
	MUSIC_STATUS old_status = mstatus;
#endif	


#ifdef USE_BACKGROUND_MUSIC
	mstatus &= ~MUSIC_SUSPENDED;
    stop_background_music(0);
#endif

    XUngrabKeyboard(disp, CurrentTime);

    if(dga_flags & DGA_DIRECT_VIDEO_ON)
	  XF86DGADirectVideo(disp, screen_no, 0);

#ifdef USE_VM_EXTENSION
    if(dga_flags & DGA_VM_CHANGED)
	{  XF86VidModeLockModeSwitch(disp, screen_no, 0);
	   XF86VidModeSwitchToMode(disp, screen_no, &modeline_original);
	}	   
#endif
    
	my_pid =  find_bossed_process();
	if (my_pid >= 0)
		 release_bossed_process (my_pid, 0);

	my_pid = getpid();   
	if (WindowIdAtom == None)
	{  fprintf (stderr, "Cant perform BOSS command: no property available\n"
												"Will terminate instead\n");
		 goto KillMe;				 
	}

    XChangeProperty(disp, root_wnd, WindowIdAtom, XA_INTEGER, 32,
			    PropModeReplace, (unsigned char *) &my_pid, 1);

    do   XNextEvent(disp, &xevnt);
    while  (xevnt.type != ClientMessage ||
			xevnt.xclient.message_type != WindowIdAtom ||
		    xevnt.xclient.data.l[0] != my_pid);

    if (xevnt.xclient.data.l[1] > 0) 
	{  

#ifdef USE_VM_EXTENSION
      if(dga_flags & DGA_VM_CHANGED)
	  {  XF86VidModeSwitchToMode(disp, screen_no, &modeline_current);
	     XF86VidModeSetViewPort(disp, screen_no, 0, 0);
		 XF86VidModeLockModeSwitch(disp, screen_no, 1);
	  }	   
#endif
      XF86DGADirectVideo(disp, screen_no, XF86DGADirectGraphics);
		LastSaveLayer = -1;
	  dga_flags |= DGA_DIRECT_VIDEO_ON;
	  
	  flush_all_pane(PT_FOREGROUND);
      XGrabKeyboard(disp, wnd, False, GrabModeAsync, GrabModeAsync, CurrentTime);

#ifdef USE_BACKGROUND_MUSIC
		mstatus |= (old_status & MUSIC_SUSPENDED);
		if (old_status & MUSIC_ON)	
 		  start_background_music();
#endif
		return;
	}

KillMe:
	  fprintf (stderr, "xifrac [pid %d] killed\n", my_pid );
	  termination_routine(SIGTERM);
	      			     
}			       

/* Mode = 0 - resume, 1 - kill */
int unboss_routine(int mode)
{   pid_t pid_bossed;
    Screen	*scr;
	int rc;

	pltf = PLTF_X11;

    disp = XOpenDisplay(NULL);
    if (disp == NULL)
    {  fprintf (stderr, "Can't open display: this is an X-Window application! \n");
       return 4;   
    }

    WindowIdAtom = XInternAtom(disp, WindowIdAtomName,  True);

     if (WindowIdAtom == (Atom)None) 
	  pid_bossed = -1;
	else	  
	{   scr = DefaultScreenOfDisplay(disp);
	    root_wnd = scr->root;
		pid_bossed = find_bossed_process();
	}		

	if (pid_bossed < 0)
	{ fprintf (stderr, "No suspended windows\n");
	  rc = 8;
	}
	else
	{ release_bossed_process (pid_bossed, mode);
	  rc = 0;
	}	  
    
	XCloseDisplay(disp);
	return 0;
} 

